@@ -1,10 +1,10 @@
 /*  Arg_parser - POSIX/GNU command line argument parser. (C++ version)
-    Copyright (C) 2006, 2007, 2008, 2009, 2010, 2011, 2012
+    Copyright (C) 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014
     Antonio Diaz Diaz.
 
     This library is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
+    the Free Software Foundation, either version 2 of the License, or
     (at your option) any later version.
 
     This library is distributed in the hope that it will be useful,
@@ -36,7 +36,7 @@
 bool Arg_parser::parse_long_option( const char * const opt, const char * const arg,
                                     const Option options[], int & argind )
   {
-  unsigned int len;
+  unsigned len;
   int index = -1;
   bool exact = false, ambig = false;
 
@@ -44,7 +44,7 @@ bool Arg_parser::parse_long_option( const char * const opt, const char * const a
 
   // Test all long options for either exact match or abbreviated matches.
   for( int i = 0; options[i].code != 0; ++i )
-    if( options[i].name && !std::strncmp( options[i].name, &opt[2], len ) )
+    if( options[i].name && std::strncmp( options[i].name, &opt[2], len ) == 0 )
       {
       if( std::strlen( options[i].name ) == len )	// Exact match found
         { index = i; exact = true; break; }
@@ -56,30 +56,30 @@ bool Arg_parser::parse_long_option( const char * const opt, const char * const a
 
   if( ambig && !exact )
     {
-    error_ = "option `"; error_ += opt; error_ += "' is ambiguous";
+    error_ = "option '"; error_ += opt; error_ += "' is ambiguous";
     return false;
     }
 
   if( index < 0 )		// nothing found
     {
-    error_ = "unrecognized option `"; error_ += opt; error_ += '\'';
+    error_ = "unrecognized option '"; error_ += opt; error_ += '\'';
     return false;
     }
 
   ++argind;
   data.push_back( Record( options[index].code ) );
 
-  if( opt[len+2] )		// `--<long_option>=<argument>' syntax
+  if( opt[len+2] )		// '--<long_option>=<argument>' syntax
     {
     if( options[index].has_arg == no )
       {
-      error_ = "option `--"; error_ += options[index].name;
+      error_ = "option '--"; error_ += options[index].name;
       error_ += "' doesn't allow an argument";
       return false;
       }
     if( options[index].has_arg == yes && !opt[len+3] )
       {
-      error_ = "option `--"; error_ += options[index].name;
+      error_ = "option '--"; error_ += options[index].name;
       error_ += "' requires an argument";
       return false;
       }
@@ -91,7 +91,7 @@ bool Arg_parser::parse_long_option( const char * const opt, const char * const a
     {
     if( !arg || !arg[0] )
       {
-      error_ = "option `--"; error_ += options[index].name;
+      error_ = "option '--"; error_ += options[index].name;
       error_ += "' requires an argument";
       return false;
       }
@@ -120,7 +120,7 @@ bool Arg_parser::parse_short_option( const char * const opt, const char * const
 
     if( index < 0 )
       {
-      error_ = "invalid option -- "; error_ += c;
+      error_ = "invalid option -- '"; error_ += c; error_ += '\'';
       return false;
       }
 
@@ -135,7 +135,8 @@ bool Arg_parser::parse_short_option( const char * const opt, const char * const
       {
       if( !arg || !arg[0] )
         {
-        error_ = "option requires an argument -- "; error_ += c;
+        error_ = "option requires an argument -- '"; error_ += c;
+        error_ += '\'';
         return false;
         }
       data.back().argument = arg; ++argind; cind = 0;
@@ -156,12 +157,12 @@ Arg_parser::Arg_parser( const int argc, const char * const argv[],
   while( argind < argc )
     {
     const unsigned char ch1 = argv[argind][0];
-    const unsigned char ch2 = ( ch1 ? argv[argind][1] : 0 );
+    const unsigned char ch2 = ch1 ? argv[argind][1] : 0;
 
     if( ch1 == '-' && ch2 )		// we found an option
       {
       const char * const opt = argv[argind];
-      const char * const arg = (argind + 1 < argc) ? argv[argind+1] : 0;
+      const char * const arg = ( argind + 1 < argc ) ? argv[argind+1] : 0;
       if( ch2 == '-' )
         {
         if( !argv[argind][2] ) { ++argind; break; }	// we found "--"
@@ -178,7 +179,7 @@ Arg_parser::Arg_parser( const int argc, const char * const argv[],
   if( error_.size() ) data.clear();
   else
     {
-    for( unsigned int i = 0; i < non_options.size(); ++i )
+    for( unsigned i = 0; i < non_options.size(); ++i )
       { data.push_back( Record() ); data.back().argument.swap( non_options[i] ); }
     while( argind < argc )
       { data.push_back( Record() ); data.back().argument = argv[argind++]; }
