@@ -1,6 +1,6 @@
 /*  GNU ddrescue - Data recovery tool
     Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012,
-    2013 Antonio Diaz Diaz.
+    2013, 2014 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -16,55 +16,30 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
-class Logbook
+class Logbook : public Logfile
   {
-public:
-  enum Status
-    { copying = '?', trimming = '*', splitting = '/', retrying = '-',
-      filling = 'F', generating = 'G', finished = '+' };
-
-private:
   const long long offset_;		// outfile offset (opos - ipos);
-  long long current_pos_;
   long long logfile_isize_;
-  Status current_status_;
   Domain & domain_;			// rescue domain
   uint8_t *iobuf_base, *iobuf_;		// iobuf is aligned to page and hardbs
-  const char * const filename_;
   const int hardbs_, softbs_;
   const char * final_msg_;
   int final_errno_;
-  mutable int index_;			// cached index of last find or change
-  std::vector< Sblock > sblock_vector;	// note: blocks are consecutive
   long ul_t1;				// variable for update_logfile
   bool logfile_exists_;
 
   Logbook( const Logbook & );		// declared as private
   void operator=( const Logbook & );	// declared as private
 
-  void erase_sblock( const int i )
-    { sblock_vector.erase( sblock_vector.begin() + i ); }
-  void insert_sblock( const int i, const Sblock & sb )
-    { sblock_vector.insert( sblock_vector.begin() + i, sb ); }
-  void split_domain_border_sblocks();
-
 public:
-  Logbook( const long long offset, const long long isize,
-           Domain & dom, const char * const logname,
-           const int cluster, const int hardbs,
-           const bool complete_only, const bool do_not_read = false );
+  Logbook( const long long offset, const long long isize, Domain & dom,
+           const char * const logname, const int cluster,
+           const int hardbs, const bool complete_only );
   ~Logbook() { delete[] iobuf_base; }
 
-  bool blank() const;
-  void compact_sblock_vector();
-  bool update_logfile( const int odes = -1, const bool force = false,
-                       const bool retry = true );
-  void write_logfile( FILE * const f ) const;
+  bool update_logfile( const int odes = -1, const bool force = false );
 
-  long long current_pos() const { return current_pos_; }
-  Status current_status() const { return current_status_; }
   const Domain & domain() const { return domain_; }
-  const char * filename() const { return filename_; }
   uint8_t * iobuf() const { return iobuf_; }
   int hardbs() const { return hardbs_; }
   int softbs() const { return softbs_; }
@@ -74,39 +49,11 @@ public:
   bool logfile_exists() const { return logfile_exists_; }
   long long logfile_isize() const { return logfile_isize_; }
 
-  void current_pos( const long long pos ) { current_pos_ = pos; }
-  void current_status( const Status st ) { current_status_ = st; }
-  void final_msg( const char * const msg ) { final_msg_ = msg; }
-  void final_errno( const int e ) { final_errno_ = e; }
+  void final_msg( const char * const msg, const int e = 0 )
+    { final_msg_ = msg; final_errno_ = e; }
 
-  const Sblock & sblock( const int i ) const
-    { return sblock_vector[i]; }
-  int sblocks() const { return (int)sblock_vector.size(); }
-  void change_sblock_status( const int i, const Sblock::Status st )
-    { sblock_vector[i].status( st ); }
-  void split_sblock_by( const long long pos, const int i )
-    {
-    if( sblock_vector[i].includes( pos ) )
-      insert_sblock( i, sblock_vector[i].split( pos ) );
-    }
-  bool truncate_vector( const long long end, const bool force = false );
   void truncate_domain( const long long end )
     { domain_.crop_by_file_size( end ); }
-
-  int find_index( const long long pos ) const;
-  int find_largest_sblock( const Sblock::Status st ) const;
-  int find_smallest_sblock( const Sblock::Status st ) const;
-  void find_chunk( Block & b, const Sblock::Status st,
-                   const int alignment = 0 ) const;
-  void rfind_chunk( Block & b, const Sblock::Status st,
-                    const int alignment = 0 ) const;
-  int change_chunk_status( const Block & b, const Sblock::Status st );
-
-  static bool isstatus( const int st )
-    { return ( st == copying || st == trimming || st == splitting ||
-               st == retrying || st == filling || st == generating ||
-               st == finished ); }
-  static const char * status_name( const Status st );
   };
 
 
@@ -122,11 +69,13 @@ class Fillbook : public Logbook
 					// variables for show_status
   long long a_rate, c_rate, first_size, last_size;
   long long last_ipos;
-  long t0, t1;
+  long t0, t1;				// start, current times
+  int oldlen;
 
   int fill_areas( const std::string & filltypes );
   int fill_block( const Block & b );
-  void show_status( const long long ipos, bool force = false );
+  void show_status( const long long ipos, const char * const msg = 0,
+                    bool force = false );
 
 public:
   Fillbook( const long long offset, Domain & dom,
@@ -136,7 +85,7 @@ public:
       ignore_write_errors_( ignore_write_errors ),
       synchronous_( synchronous ),
       a_rate( 0 ), c_rate( 0 ), first_size( 0 ), last_size( 0 ),
-      last_ipos( 0 ), t0( 0 ), t1( 0 )
+      last_ipos( 0 ), t0( 0 ), t1( 0 ), oldlen( 0 )
       {}
 
   int do_fill( const int odes, const std::string & filltypes );
@@ -151,7 +100,7 @@ class Genbook : public Logbook
 					// variables for show_status
   long long a_rate, c_rate, first_size, last_size;
   long long last_ipos;
-  long t0, t1;
+  long t0, t1;				// start, current times
   int oldlen;
 
   void check_block( const Block & b, int & copied_size, int & error_size );
@@ -173,7 +122,7 @@ public:
 
 struct Rb_options
   {
-  enum { default_skipbs = 65536, max_skipbs = 1 << 30 };
+  enum { default_skipbs = 65536, max_max_skipbs = 1 << 30 };
 
   long long max_error_rate;
   long long min_outfile_size;
@@ -182,19 +131,26 @@ struct Rb_options
   int max_errors;
   int max_logfile_size;
   int max_retries;
+  int o_direct;			// O_DIRECT or 0
   int skipbs;			// initial size to skip on read error
+  int max_skipbs;		// maximum size to skip on read error
   bool complete_only;
   bool new_errors_only;
   bool nosplit;
+  bool notrim;
+  bool reopen_on_error;
   bool retrim;
+  bool reverse;
   bool sparse;
   bool try_again;
 
   Rb_options()
     : max_error_rate( -1 ), min_outfile_size( -1 ), min_read_rate( -1 ),
-      timeout( -1 ), max_errors( -1 ), max_logfile_size( 1000 ),
-      max_retries( 0 ), skipbs( default_skipbs ), complete_only( false ),
-      new_errors_only( false ), nosplit( false ), retrim( false ),
+      timeout( -1 ), max_errors( -1 ), max_logfile_size( 10000 ),
+      max_retries( 0 ), o_direct( 0 ), skipbs( default_skipbs ),
+      max_skipbs( max_max_skipbs ), complete_only( false ),
+      new_errors_only( false ), nosplit( false ), notrim( false ),
+      reopen_on_error( false ), retrim( false ), reverse( false ),
       sparse( false ), try_again( false )
       {}
 
@@ -204,10 +160,13 @@ struct Rb_options
                min_read_rate == o.min_read_rate && timeout == o.timeout &&
                max_errors == o.max_errors &&
                max_logfile_size == o.max_logfile_size &&
-               max_retries == o.max_retries && skipbs == o.skipbs &&
+               max_retries == o.max_retries && o_direct == o.o_direct &&
+               skipbs == o.skipbs && max_skipbs == o.max_skipbs &&
                complete_only == o.complete_only &&
                new_errors_only == o.new_errors_only &&
-               nosplit == o.nosplit && retrim == o.retrim &&
+               nosplit == o.nosplit && notrim == o.notrim &&
+               reopen_on_error == o.reopen_on_error &&
+               retrim == o.retrim && reverse == o.reverse &&
                sparse == o.sparse && try_again == o.try_again ); }
   bool operator!=( const Rb_options & o ) const
     { return !( *this == o ); }
@@ -219,19 +178,20 @@ class Rescuebook : public Logbook, public Rb_options
   long long error_rate;
   long long sparse_size;		// end position of pending writes
   long long recsize, errsize;		// total recovered and error sizes
+  const Domain * const test_domain;	// good/bad map for test mode
   const char * const iname_;
-  const int max_skip_size;		// maximum size to skip on read error
   int e_code;				// error code for errors_or_timeout
 					// 1 rate, 2 errors, 4 timeout
   int errors;				// error areas found so far
   int ides_, odes_;			// input and output file descriptors
-  const bool synchronous_;
+  const bool access_works, synchronous_;
 					// variables for update_rates
   long long a_rate, c_rate, first_size, last_size;
   long long last_ipos;
   long t0, t1, ts;			// start, current, last successful
   int oldlen;
   bool rates_updated;
+  bool first_post;			// variable for show_status
 
   bool extend_outfile_size();
   int copy_block( const Block & b, int & copied_size, int & error_size );
@@ -242,18 +202,23 @@ class Rescuebook : public Logbook, public Rb_options
   void reduce_min_read_rate()
     { if( min_read_rate > 0 ) min_read_rate /= 10; }
   bool slow_read() const
-    { return ( t1 - t0 >= 10 &&		// no slow reads for first 10s
+    { return ( t1 - t0 >= 30 &&		// no slow reads for first 30s
                ( ( min_read_rate > 0 && c_rate < min_read_rate &&
                    c_rate < a_rate / 2 ) ||
                  ( min_read_rate == 0 && c_rate < a_rate / 10 ) ) ); }
-  int copy_and_update( const Block & b, const Sblock::Status st,
-                       int & copied_size, int & error_size,
-                       const char * const msg, bool & first_post,
-                       const bool forward );
+  int update( const Block & b, const Sblock::Status st,
+              const int copied_size, const int error_size );
+  int copy_and_update( const Block & b, int & error_size,
+                       const char * const msg, const bool forward );
+  int copy_and_update2( const Block & b, int & copied_size,
+                        int & error_size, const char * const msg,
+                        const bool forward, const bool small_try );
+  bool reopen_infile();
   int copy_non_tried();
-  int rcopy_non_tried();
+  int fcopy_non_tried( const char * const msg, const int pass );
+  int rcopy_non_tried( const char * const msg, const int pass );
   int trim_errors();
-  int split_errors( const bool reverse );
+  int split_errors();
   int copy_errors();
   int rcopy_errors();
   void update_rates( const bool force = false );
@@ -261,12 +226,12 @@ class Rescuebook : public Logbook, public Rb_options
                     const bool force = false );
 public:
   Rescuebook( const long long offset, const long long isize,
-              Domain & dom, const Rb_options & rb_opts,
-              const char * const iname, const char * const logname,
-              const int cluster, const int hardbs,
-              const bool synchronous );
+              Domain & dom, const Domain * const test_dom,
+              const Rb_options & rb_opts, const char * const iname,
+              const char * const logname, const int cluster,
+              const int hardbs, const bool synchronous );
 
-  int do_rescue( const int ides, const int odes, const bool reverse );
+  int do_rescue( const int ides, const int odes );
   };
 
 
@@ -288,14 +253,4 @@ inline int round_up( int size, const int hardbs )
 const char * format_time( long t );
 bool interrupted();
 void set_signals();
-
-
-// Defined in main_common.cc
-//
-extern int verbosity;
-void internal_error( const char * const msg );
-void show_error( const char * const msg,
-                 const int errcode = 0, const bool help = false );
-void write_logfile_header( FILE * const f );
-const char * format_num( long long num, long long limit = 999999,
-                         const int set_prefix = 0 );
+int signaled_exit();
