@@ -1,6 +1,6 @@
 /*  GNU ddrescue - Data recovery tool
-    Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010
-    Antonio Diaz Diaz.
+    Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012,
+    2013, 2014 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -18,10 +18,14 @@
 
 #define _FILE_OFFSET_BITS 64
 
+#include <algorithm>
 #include <climits>
 #include <cstdio>
+#include <cstring>
+#include <ctime>
 #include <string>
 #include <vector>
+#include <stdint.h>
 
 #include "block.h"
 #include "ddrescue.h"
@@ -31,6 +35,7 @@
 //
 int Fillbook::fill_areas( const std::string & filltypes )
   {
+  const char * const msg = "Filling blocks...";
   bool first_post = true;
 
   for( int index = 0; index < sblocks(); ++index )
@@ -39,16 +44,23 @@ int Fillbook::fill_areas( const std::string & filltypes )
     if( !domain().includes( sb ) ) { if( domain() < sb ) break; else continue; }
     if( sb.end() <= current_pos() ||
         filltypes.find( sb.status() ) >= filltypes.size() ) continue;
-    Block b( sb.pos(), softbs() );
+    Block b( sb.pos(), softbs() );	// fill the area a softbs at a time
     if( sb.includes( current_pos() ) ) b.pos( current_pos() );
     if( b.end() > sb.end() ) b.crop( sb );
-    current_status( filling );
+    current_status( filling, msg );
     while( b.size() > 0 )
       {
+      current_pos( b.pos() );
       if( verbosity >= 0 )
-        { show_status( b.pos(), first_post ); first_post = false; }
+        { show_status( b.pos(), msg, first_post ); first_post = false; }
+      if( interrupted() ) return -1;
       const int retval = fill_block( b );
-      if( retval ) return retval;
+      if( retval )					// write error
+        {
+        if( !ignore_write_errors_ ) return retval;
+        if( b.size() > hardbs() )	// fill the area a hardbs at a time
+          { b.size( hardbs() ); continue; }
+        }
       if( !update_logfile( odes_ ) ) return -2;
       b.pos( b.end() );
       if( b.end() > sb.end() ) b.crop( sb );
@@ -59,6 +71,54 @@ int Fillbook::fill_areas( const std::string & filltypes )
   }
 
 
+void Fillbook::show_status( const long long ipos, const char * const msg,
+                            bool force )
+  {
+  const char * const up = "\x1b[A";
+  if( t0 == 0 )
+    {
+    t0 = t1 = initial_time();
+    first_size = last_size = filled_size;
+    force = true;
+    std::printf( "\n\n\n" );
+    }
+
+  if( ipos >= 0 ) last_ipos = ipos;
+  const long t2 = std::time( 0 );
+  if( t2 > t1 || force )
+    {
+    if( t2 > t1 )
+      {
+      a_rate = ( filled_size - first_size ) / ( t2 - t0 );
+      c_rate = ( filled_size - last_size ) / ( t2 - t1 );
+      t1 = t2;
+      last_size = filled_size;
+      }
+    std::printf( "\r%s%s%s", up, up, up );
+    std::printf( "filled size: %10sB,  filled areas: %6u,  current rate: %9sB/s\n",
+                 format_num( filled_size ), filled_areas,
+                 format_num( c_rate, 99999 ) );
+    std::printf( "remain size: %10sB,  remain areas: %6u,  average rate: %9sB/s\n",
+                 format_num( remaining_size ), remaining_areas,
+                 format_num( a_rate, 99999 ) );
+    std::printf( "current pos: %10sB,  run time:  %9s\n",
+                 format_num( last_ipos + offset() ), format_time( t1 - t0 ) );
+    if( msg && msg[0] )
+      {
+      const int len = std::strlen( msg ); std::printf( "\r%s", msg );
+      for( int i = len; i < oldlen; ++i ) std::fputc( ' ', stdout );
+      oldlen = len;
+      }
+    std::fflush( stdout );
+    }
+  else if( t2 < t1 )			// clock jumped back
+    {
+    t0 -= std::min( t0, t1 - t2 );
+    t1 = t2;
+    }
+  }
+
+
 // Return values: 1 write error, 0 OK.
 //
 int Fillbook::do_fill( const int odes, const std::string & filltypes )
@@ -86,33 +146,35 @@ int Fillbook::do_fill( const int odes, const std::string & filltypes )
   if( verbosity >= 0 )
     {
     std::printf( "Press Ctrl-C to interrupt\n" );
-    if( filename() )
+    if( logfile_exists() )
       {
       std::printf( "Initial status (read from logfile)\n" );
-      std::printf( "filled size:    %10sB,", format_num( filled_size ) );
-      std::printf( "  filled areas:    %7u\n", filled_areas );
-      std::printf( "remaining size: %10sB,", format_num( remaining_size ) );
-      std::printf( "  remaining areas: %7u\n", remaining_areas );
+      std::printf( "filled size:    %10sB,  filled areas:    %7u\n",
+                   format_num( filled_size ), filled_areas );
+      std::printf( "remaining size: %10sB,  remaining areas: %7u\n",
+                   format_num( remaining_size ), remaining_areas );
       std::printf( "Current status\n" );
       }
     }
   int retval = fill_areas( filltypes );
+  const bool signaled = ( retval == -1 );
+  if( signaled ) retval = 0;
   if( verbosity >= 0 )
     {
-    show_status( -1, true );
-    if( retval == 0 ) std::printf( "Finished" );
-    else if( retval == -2 ) std::printf( "Logfile error" );
-    else if( retval < 0 ) std::printf( "Interrupted by user" );
+    show_status( -1, ( retval || signaled ) ? 0 : "Finished", true );
+    if( retval == -2 ) std::printf( "\nLogfile error" );
+    else if( signaled ) std::printf( "\nInterrupted by user" );
     std::fputc( '\n', stdout );
     }
   if( retval == -2 ) retval = 1;		// logfile error
   else
     {
-    if( retval == 0 ) current_status( finished );
-    else if( retval < 0 ) retval = 0;		// interrupted by user
+    if( retval == 0 && !signaled ) current_status( finished );
     compact_sblock_vector();
     if( !update_logfile( odes_, true ) && retval == 0 ) retval = 1;
     }
   if( final_msg() ) show_error( final_msg(), final_errno() );
-  return retval;
+  if( retval ) return retval;		// errors have priority over signals
+  if( signaled ) return signaled_exit();
+  return 0;
   }
